###############################
###  TETRIS v1.01 for TML   ###
###   by Piotr Kowalewski   ###
### This is a demonstration ###
###       game using        ###
###     the tml module      ###
###      for HP Prime       ###
###    October 6, 2024      ###
###############################
import tml
import urandom
from hpprime import eval as ppleval

# Initialize terminal with 5x10 pixel font
t = tml.tml(font='std5x10', status=None)
t.clear()

# Game parameters
WIDTH = const(10)  # Width of the board
HEIGHT = const(20)  # Height of the board
BLOCK = '[]'  # Block representation (two characters)
EMPTY = '  '  # Empty space on the board (two spaces)
BORDER_LEFT = '|'  # Left border
BORDER_RIGHT = '|'  # Right border
BORDER_BOTTOM = '='  # Bottom border
SCORE_X = const(WIDTH * 2 + 4)  # X position for score display

BOARD = [[EMPTY for _ in range(WIDTH)] for _ in range(HEIGHT)]

# Define Tetris shapes
SHAPES = [
    # I
    [[BLOCK, BLOCK, BLOCK, BLOCK]],
    # O
    [[BLOCK, BLOCK],
     [BLOCK, BLOCK]],
    # T
    [[BLOCK, BLOCK, BLOCK],
     [EMPTY, BLOCK, EMPTY]],
    # S
    [[EMPTY, BLOCK, BLOCK],
     [BLOCK, BLOCK, EMPTY]],
    # Z
    [[BLOCK, BLOCK, EMPTY],
     [EMPTY, BLOCK, BLOCK]],
    # J
    [[BLOCK, EMPTY, EMPTY],
     [BLOCK, BLOCK, BLOCK]],
    # L
    [[EMPTY, EMPTY, BLOCK],
     [BLOCK, BLOCK, BLOCK]],
]

def draw_board(temp_board, score):
    t.set_cursor(0, 0)
    # Draw each row with borders
    for y, row in enumerate(temp_board):
        line = BORDER_LEFT + ''.join(row) + BORDER_RIGHT
        t.print(line)
    # Draw bottom border
    t.print(BORDER_LEFT + BORDER_BOTTOM * (WIDTH * 2) + BORDER_RIGHT)
    # Draw score on the right side
    t.set_cursor(SCORE_X, 4)
    t.print('Score: %d' % score)
    #t.set_cursor(SCORE_X, 4)
    #t.print(str(score))

def check_collision(shape, offset_x, offset_y):
    for y, row in enumerate(shape):
        for x, cell in enumerate(row):
            if cell == BLOCK:
                new_x = x + offset_x
                new_y = y + offset_y
                # Check boundaries and existing blocks
                if new_x < 0 or new_x >= WIDTH or new_y >= HEIGHT:
                    return True
                if new_y >= 0 and BOARD[new_y][new_x] == BLOCK:
                    return True
    return False

def merge_shape(shape, offset_x, offset_y):
    # Merge the current shape into the board
    for y, row in enumerate(shape):
        for x, cell in enumerate(row):
            if cell == BLOCK:
                BOARD[y + offset_y][x + offset_x] = BLOCK

def remove_full_lines():
    global BOARD
    lines_cleared = 0
    new_board = []
    # Remove full lines and count them
    for row in BOARD:
        if all(cell == BLOCK for cell in row):
            lines_cleared += 1
        else:
            new_board.append(row)
    # Add empty lines on top
    for _ in range(lines_cleared):
        new_board.insert(0, [EMPTY for _ in range(WIDTH)])
    BOARD = new_board
    return lines_cleared

def rotate_shape(shape):
    # Rotate the shape clockwise
    return [list(row) for row in zip(*shape[::-1])]

previous_time = ppleval("time()")  # current time in seconds

def game_loop():
    score = 0
    current_shape = urandom.choice(SHAPES)
    offset_x = WIDTH // 2 - len(current_shape[0]) // 2
    offset_y = 0
    gravity_delay = 0.5  # Gravity delay in seconds
    gravity_timer = 0

    previous_time = ppleval("time()")  # Initial time
    last_move_time = previous_time  # Time of the last left/right movement
    move_delay = 0.2  # Delay for horizontal movement (in seconds)

    rotation_handled = False  # Control rotation handling
    t.print_xy(SCORE_X, 1, 'TETRIS v1.01 for TML')
    t.print_xy(SCORE_X, 2, 'by Piotr Kowalewski (komame)')

    while True:
        current_time = ppleval("time()")  # Get the current time
        delta_time = current_time - previous_time  # Calculate time difference
        previous_time = current_time  # Update the previous time
        gravity_timer += delta_time  # Update gravity timer

        # Input handling
        keys = t.get_keys()
        if keys:
            key = keys[0]
            if key == 4:  # Esc key to exit
                break
            elif (key == 7 or key == 8) and (current_time - last_move_time >= move_delay):  # Left/Right movement
                if key == 7:  # Left arrow key
                    if not check_collision(current_shape, offset_x - 1, offset_y):
                        offset_x -= 1
                elif key == 8:  # Right arrow key
                    if not check_collision(current_shape, offset_x + 1, offset_y):
                        offset_x += 1
                last_move_time = current_time  # Update time of the last movement
            elif key == 12:  # Down arrow key
                if not check_collision(current_shape, offset_x, offset_y + 1):
                    offset_y += 1
                    gravity_timer = 0  # Reset gravity timer
            elif key == 2:  # Up arrow key (Rotate)
                if not rotation_handled:
                    rotated_shape = rotate_shape(current_shape)
                    if not check_collision(rotated_shape, offset_x, offset_y):
                        current_shape = rotated_shape
                    rotation_handled = True
        else:
            rotation_handled = False  # Reset rotation when no key is pressed

        # Gravity effect
        if gravity_timer >= gravity_delay:  # Apply gravity after delay
            gravity_timer = 0
            if check_collision(current_shape, offset_x, offset_y + 1):
                merge_shape(current_shape, offset_x, offset_y)
                score += 1  # Add 1 point for each dropped block
                lines_cleared = remove_full_lines()
                if lines_cleared > 0:
                    score += lines_cleared * 10  # Add 10 points per line cleared
                # Spawn a new shape
                current_shape = urandom.choice(SHAPES)
                offset_x = WIDTH // 2 - len(current_shape[0]) // 2
                offset_y = 0
                if check_collision(current_shape, offset_x, offset_y):
                    # Game over condition
                    t.set_cursor(SCORE_X, 6)
                    t.print('GAME OVER')
                    break
            else:
                offset_y += 1

        # Draw the board with the current shape
        temp_board = [row[:] for row in BOARD]
        for y, row in enumerate(current_shape):
            for x, cell in enumerate(row):
                if cell == BLOCK:
                    new_x = x + offset_x
                    new_y = y + offset_y
                    if 0 <= new_x < WIDTH and 0 <= new_y < HEIGHT:
                        temp_board[new_y][new_x] = BLOCK

        draw_board(temp_board, score)

    t.set_cursor(SCORE_X, 20)
    t.print('\n\nPress any key to exit.')
    t.read_key()

# Start the game
game_loop()
